import open3d as o3d
from typing import List, Tuple, Dict, Any
import numpy as np

def pointcloud_audit(pcd: o3d.geometry.PointCloud)-> Dict[str, Any]:
    '''
    Audit point cloud.

    :param pcd: Point cloud.
    :type pcd: o3d.geometry.PointCloud
    :return: Dictionary of audit results.
    :rtype: Dict[str, Any]

    :Example:

    ::
    '''
    dict_results = {}
    number_pcd_points = len(pcd.points)
    dict_results["number_pcd_points"] = number_pcd_points
    dict_results["has_normals"] = pcd.has_normals()
    dict_results["has_colors"] = pcd.has_colors()
    dict_results["is_empty"] = pcd.is_empty()
    if not pcd.is_empty():
        # get maximum and minimum values
        np_points = np.asarray(pcd.points)
        dict_results["max_x"] = np.max(np_points[:, 0])
        dict_results["min_x"] = np.min(np_points[:, 0])
        dict_results["max_y"] = np.max(np_points[:, 1])
        dict_results["min_y"] = np.min(np_points[:, 1])
        dict_results["max_z"] = np.max(np_points[:, 2])
        dict_results["min_z"] = np.min(np_points[:, 2])
        # check if all points are finite
        dict_results["all_points_finite"] = np.all(np.isfinite(np_points))
        # check if all points are unique
        dict_results["all_points_unique"] = len(np_points) == len(np.unique(np_points, axis=0))
    return dict_results

def pointcloud_sanitize(pcd: o3d.geometry.PointCloud)->o3d.geometry.PointCloud:
    # check if the pointcloud is empty and then remove nonfinite and duplicate points
    if pcd.is_empty():
        return pcd
    # remove non finite points and associated colors
    pcd_has_colors = pcd.has_colors()
    np_points = np.asarray(pcd.points)
    np_points = np_points[np.isfinite(np_points).all(axis=1)]
    if pcd_has_colors:
        np_colors = np.asarray(pcd.colors)
        np_colors = np_colors[np.isfinite(np_points).all(axis=1)]
    # remove duplicate points and associated colors
    _, unique_indices = np.unique(np_points, axis=0, return_index=True)
    pcd.points = o3d.utility.Vector3dVector(np_points[unique_indices])
    if pcd_has_colors:
        pcd.colors = o3d.utility.Vector3dVector(np_colors[unique_indices])
    return pcd

"""
pcd = o3d.geometry.PointCloud()
dict_results = pointcloud_audit(pcd)
print(dict_results)
pcd.points = o3d.utility.Vector3dVector(np.random.randn(100, 3))
dict_results = pointcloud_audit(pcd)
print(dict_results)
pcd_filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
pcd = o3d.io.read_point_cloud(pcd_filename)
dict_results = pointcloud_audit(pcd)
print(dict_results)
o3d.visualization.draw_geometries([pcd])
pcd = pointcloud_sanitize(pcd)
dict_results = pointcloud_audit(pcd)
print(dict_results)
o3d.visualization.draw_geometries([pcd])
"""


import os
import glob
import pickle

def normalize_values(dictionary):
    max_value = dictionary['maximum']
    normalized_dict = {key: value / max_value for key, value in dictionary.items() if isinstance(value, (int, float))}
    return normalized_dict

def create_percentage_files(directory):
    for filepath in glob.glob(directory + '**/*_means.pkl', recursive=True):
        with open(filepath, 'rb') as file:
            data = pickle.load(file)
        normalized_data = normalize_values(data)
        new_filepath = filepath.replace('_means.pkl', '_percentages.pkl')
        with open(new_filepath, 'wb') as file:
            pickle.dump(normalized_data, file)

directory = "/home/scpmaotj/pkl_files/me/"  # specify your directory
create_percentage_files(directory)



import numpy as np
import pandas as pd
from sklearn.cluster import KMeans, DBSCAN, AgglomerativeClustering
import matplotlib.pyplot as plt

def run_clustering(algorithm, data, **kwargs):
    # Initialize the clustering algorithm
    clustering = algorithm(**kwargs)
    
    # Perform clustering on the data
    labels = clustering.fit_predict(data)
    
    # Return the cluster labels
    return labels

def clustering_report(data):
    # Perform clustering with different algorithms and parameters
    kmeans_labels = run_clustering(KMeans, data, n_clusters=3)
    dbscan_labels = run_clustering(DBSCAN, data, eps=0.5, min_samples=5)
    agg_labels = run_clustering(AgglomerativeClustering, data, n_clusters=3)
    
    # Create a DataFrame to store the results
    results_df = pd.DataFrame({"KMeans": kmeans_labels, "DBSCAN": dbscan_labels, "Agglomerative": agg_labels})
    
    # Generate descriptive statistics for each cluster algorithm
    report = results_df.describe()
    
    # Print the report
    return results_df, report





import mrdja.ransac.coreransac as coreransac
import mrdja.ransac.coreransaccuda as coreransaccuda
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geom
import mrdja.sampling as sampling
import open3d as o3d
import numpy as np
import random
from typing import Callable, List, Tuple, Union, Dict
from numba import cuda

def get_ransac_data_from_np_points(np_points: np.ndarray, ransac_iterator: Callable, ransac_iterations: int = 100, threshold: float = 0.1, 
                                  verbosity_level: int = 0, inherited_verbose_string: str = "",
                                  seed: int = None) -> Dict:
    '''
    Gets the ransac data from a file. 
    
    The ransac data is a dictionary with the following keys:
    - ransac_iterations_results: the results of the ransac iterations. It is a list, where each element is a dictionary with the following keys:
        - current_random_points: the current random points chosen to estimate the line
        - current_line: the current line estimated from the random points. Right now it is a numpy array with two points, the same as the current_random_points
        - threshold: the maximum distance from the line to consider a point as an inlier
        - number_inliers: the number of inliers
        - indices_inliers: the indices of the inliers in the point cloud
    - ransac_best_iteration_results: the best iteration results

    :param np_points: The point cloud data.
    :type np_points: np.ndarray
    :param ransac_iterator: The ransac iterator.
    :type ransac_iterator: Callable, function that takes the point cloud data and returns the ransac data.
    :param ransac_iterations: The number of ransac iterations.
    :type ransac_iterations: int
    :param threshold: The threshold.
    :type threshold: float
    :param audit_cloud: Whether to audit the cloud.
    :type audit_cloud: bool
    :param verbosity_level: The verbosity level.
    :type verbosity_level: int
    :param inherited_verbose_string: The inherited verbose string.
    :type inherited_verbose_string: str
    :param seed: The seed.
    :type seed: int
    :return: The ransac data.
    :rtype: dict

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterator = coreransac.get_ransac_line_iteration_results
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> pcd = o3d.io.read_point_cloud(office_filename)
        >>> np_points = np.asarray(pcd.points)
        >>> ransac_data = ransaclp.get_ransac_data_from_filename(np_points, ransac_iterator = ransac_iterator, 
                                                           ransac_iterations = ransac_iterations, 
                                                           threshold = threshold, audit_cloud=True, seed = seed)
        >>> iterations_results = ransac_data["ransac_iterations_results"]
        >>> len(iterations_results)
        200
        >>> first_iteration_results = iterations_results[0]
        >>> first_iteration_results["current_random_points"]
        array([[1.88671875, 1.96484375, 1.91060746],
        [3.18359375, 1.89562058, 2.45703125]])
        >>> first_iteration_results["current_line"]
        array([[1.88671875, 1.96484375, 1.91060746],
        [3.18359375, 1.89562058, 2.45703125]])
        >>> first_iteration_results["threshold"]
        0.02
        >>> first_iteration_results["number_inliers"]
        203
        >>> first_iteration_results["indices_inliers"][:5]
        array([ 58884,  59966,  60516,  61070, 138037])
        >>> best_iteration_results = ransac_data["ransac_best_iteration_results"]
        >>> best_iteration_results["current_random_points"]
        array([[1.85546875, 2.67396379, 2.08203125],
        [0.85546875, 2.57869077, 2.52734375]])
        >>> best_iteration_results["current_line"]
        array([[1.85546875, 2.67396379, 2.08203125],
        [0.85546875, 2.57869077, 2.52734375]])
        >>> best_iteration_results["threshold"]
        0.02
        >>> best_iteration_results["number_inliers"]
        2128
        >>> best_iteration_results["indices_inliers"][:5]
        array([24335, 25743, 25746, 25897, 26405])
    '''
    if seed is not None:
        np.random.seed(seed)

    dict_full_results = {}
    dict_full_results["ransac_iterations_results"] = []

    max_number_inliers = 0
    best_iteration_results = None
    current_iteration = 0
    for i in range(ransac_iterations):
        current_iteration += 1
        message = f"{inherited_verbose_string} Fitting line to pointcloud: Iteration {current_iteration}/{ransac_iterations}"
        if (verbosity_level == 1 and current_iteration % 10 == 0) or verbosity_level >= 2:
            print(message)
        else:
            dict_iteration_results = ransac_iterator(np_points, threshold, len(np_points))
        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results

def get_ransac_data_from_np_points_cuda(np_points: np.ndarray, ransac_iterator: Callable, ransac_iterations: int = 100, threshold: float = 0.1, 
                                  verbosity_level: int = 0, inherited_verbose_string: str = "",
                                  seed: int = None) -> Dict:
    '''
    Gets the ransac data from a file. 
    
    The ransac data is a dictionary with the following keys:
    - ransac_iterations_results: the results of the ransac iterations. It is a list, where each element is a dictionary with the following keys:
        - current_random_points: the current random points chosen to estimate the line
        - current_line: the current line estimated from the random points. Right now it is a numpy array with two points, the same as the current_random_points
        - threshold: the maximum distance from the line to consider a point as an inlier
        - number_inliers: the number of inliers
        - indices_inliers: the indices of the inliers in the point cloud
    - ransac_best_iteration_results: the best iteration results

    :param np_points: The point cloud data.
    :type np_points: np.ndarray
    :param ransac_iterator: The ransac iterator.
    :type ransac_iterator: Callable, function that takes the point cloud data and returns the ransac data.
    :param ransac_iterations: The number of ransac iterations.
    :type ransac_iterations: int
    :param threshold: The threshold.
    :type threshold: float
    :param audit_cloud: Whether to audit the cloud.
    :type audit_cloud: bool
    :param verbosity_level: The verbosity level.
    :type verbosity_level: int
    :param inherited_verbose_string: The inherited verbose string.
    :type inherited_verbose_string: str
    :param seed: The seed.
    :type seed: int
    :return: The ransac data.
    :rtype: dict

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterator = coreransac.get_ransac_line_iteration_results
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> pcd = o3d.io.read_point_cloud(office_filename)
        >>> np_points = np.asarray(pcd.points)
        >>> ransac_data = ransaclp.get_ransac_data_from_filename(np_points, ransac_iterator = ransac_iterator, 
                                                           ransac_iterations = ransac_iterations, 
                                                           threshold = threshold, audit_cloud=True, seed = seed)
        >>> iterations_results = ransac_data["ransac_iterations_results"]
        >>> len(iterations_results)
        200
        >>> first_iteration_results = iterations_results[0]
        >>> first_iteration_results["current_random_points"]
        array([[1.88671875, 1.96484375, 1.91060746],
        [3.18359375, 1.89562058, 2.45703125]])
        >>> first_iteration_results["current_line"]
        array([[1.88671875, 1.96484375, 1.91060746],
        [3.18359375, 1.89562058, 2.45703125]])
        >>> first_iteration_results["threshold"]
        0.02
        >>> first_iteration_results["number_inliers"]
        203
        >>> first_iteration_results["indices_inliers"][:5]
        array([ 58884,  59966,  60516,  61070, 138037])
        >>> best_iteration_results = ransac_data["ransac_best_iteration_results"]
        >>> best_iteration_results["current_random_points"]
        array([[1.85546875, 2.67396379, 2.08203125],
        [0.85546875, 2.57869077, 2.52734375]])
        >>> best_iteration_results["current_line"]
        array([[1.85546875, 2.67396379, 2.08203125],
        [0.85546875, 2.57869077, 2.52734375]])
        >>> best_iteration_results["threshold"]
        0.02
        >>> best_iteration_results["number_inliers"]
        2128
        >>> best_iteration_results["indices_inliers"][:5]
        array([24335, 25743, 25746, 25897, 26405])
    '''
    if seed is not None:
        np.random.seed(seed)

    dict_full_results = {}
    dict_full_results["ransac_iterations_results"] = []

    max_number_inliers = 0
    best_iteration_results = None
    current_iteration = 0
    # Extract x, y, z coordinates from np_points
    points_x = np_points[:, 0]
    points_y = np_points[:, 1]
    points_z = np_points[:, 2]
    # Convert points_x, points_y, and points_z to contiguous arrays
    d_points_x = np.ascontiguousarray(points_x)
    d_points_y = np.ascontiguousarray(points_y)
    d_points_z = np.ascontiguousarray(points_z)
    d_points_x = cuda.to_device(d_points_x)
    d_points_y = cuda.to_device(d_points_y)
    d_points_z = cuda.to_device(d_points_z)
    random_numbers_pairs = sampling.sampling_np_arrays_from_enumerable(np_points, cardinality_of_np_arrays=2, 
                                                                       number_of_np_arrays=ransac_iterations, 
                                                                       num_source_elems=len(np_points), seed=seed)

    for i in range(ransac_iterations):
        current_iteration += 1
        message = f"{inherited_verbose_string} Fitting line to pointcloud: Iteration {current_iteration}/{ransac_iterations}"
        if (verbosity_level == 1 and current_iteration % 10 == 0) or verbosity_level >= 2:
            print(message)

        random_points = random_numbers_pairs[i]
        dict_iteration_results = ransac_iterator(np_points, d_points_x, d_points_y, d_points_z, threshold, random_points)

        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results


def get_ransac_data_from_filename(filename: str, ransac_iterator: Callable, ransac_iterations: int = 100, threshold: float = 0.1, 
                                  audit_cloud: bool = False, verbosity_level: int = 0, inherited_verbose_string: str = "",
                                  seed: int = None) -> Dict:
    '''
    Gets the ransac data from a file. 
    
    The ransac data is a dictionary with the following keys:
    - filename: the filename
    - audit_before_sanitizing: the audit of the point cloud before sanitizing; it is a dictionary with the following keys:
        - number_pcd_points: the number of points in the point cloud
        - has_normals: whether the point cloud has normals
        - has_colors: whether the point cloud has colors
        - is_empty: whether the point cloud is empty
        - max_x: the maximum x coordinate
        - min_x: the minimum x coordinate
        - max_y: the maximum y coordinate
        - min_y: the minimum y coordinate
        - max_z: the maximum z coordinate
        - min_z: the minimum z coordinate
        - all_points_finite: whether all the points are finite
        - all_points_unique: whether all the points are unique
    - audit_after_sanitizing: the audit of the point cloud after sanitizing
    - number_pcd_points: the number of points in the point cloud
    - ransac_iterations_results: the results of the ransac iterations. It is a list, where each element is a dictionary with the following keys:
        - current_random_points: the current random points chosen to estimate the line
        - current_line: the current line estimated from the random points. Right now it is a numpy array with two points, the same as the current_random_points
        - threshold: the maximum distance from the line to consider a point as an inlier
        - number_inliers: the number of inliers
        - indices_inliers: the indices of the inliers in the point cloud
    - ransac_best_iteration_results: the best iteration results

    :param filename: The filename.
    :type filename: str
    :param ransac_iterator: The ransac iterator.
    :type ransac_iterator: Callable, function that takes the point cloud data and returns the ransac data.
    :param ransac_iterations: The number of ransac iterations.
    :type ransac_iterations: int
    :param threshold: The threshold.
    :type threshold: float
    :param audit_cloud: Whether to audit the cloud.
    :type audit_cloud: bool
    :param verbosity_level: The verbosity level.
    :type verbosity_level: int
    :param inherited_verbose_string: The inherited verbose string.
    :type inherited_verbose_string: str
    :param seed: The seed.
    :type seed: int
    :return: The ransac data.
    :rtype: dict

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterator = coreransac.get_ransac_line_iteration_results
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> ransac_data = ransaclp.get_ransac_data_from_filename(office_filename, ransac_iterator = ransac_iterator, 
                                                           ransac_iterations = ransac_iterations, 
                                                           threshold = threshold, audit_cloud=True, seed = seed)
        >>> ransac_data["filename"]
        '/home/user/open3d_data/extract/OfficePointClouds/cloud_bin_0.ply'
        >>> ransac_data["audit_before_sanitizing"]
        {'number_pcd_points': 276871,
        'has_normals': True,
        'has_colors': True,
        'is_empty': False,
        'max_x': 3.5121824741363525,
        'min_x': 0.00390625,
        'max_y': 2.80859375,
        'min_y': 0.47265625,
        'max_z': 2.7512423992156982,
        'min_z': 0.94921875,
        'all_points_finite': True,
        'all_points_unique': True}
        >>> ransac_data["audit_after_sanitizing"]
        {'number_pcd_points': 276871,
        'has_normals': True,
        'has_colors': True,
        'is_empty': False,
        'max_x': 3.5121824741363525,
        'min_x': 0.00390625,
        'max_y': 2.80859375,
        'min_y': 0.47265625,
        'max_z': 2.7512423992156982,
        'min_z': 0.94921875,
        'all_points_finite': True,
        'all_points_unique': True}
        >>> ransac_data["number_pcd_points"]
        276871
        >>> iterations_results = ransac_data["ransac_iterations_results"]
        >>> len(iterations_results)
        200
        >>> first_iteration_results = iterations_results[0]
        >>> first_iteration_results["current_random_points"]
        array([[1.88671875, 1.96484375, 1.91060746],
        [3.18359375, 1.89562058, 2.45703125]])
        >>> first_iteration_results["current_line"]
        array([[1.88671875, 1.96484375, 1.91060746],
        [3.18359375, 1.89562058, 2.45703125]])
        >>> first_iteration_results["threshold"]
        0.02
        >>> first_iteration_results["number_inliers"]
        203
        >>> first_iteration_results["indices_inliers"][:5]
        array([ 58884,  59966,  60516,  61070, 138037])
        >>> best_iteration_results = ransac_data["ransac_best_iteration_results"]
        >>> best_iteration_results["current_random_points"]
        array([[1.85546875, 2.67396379, 2.08203125],
        [0.85546875, 2.57869077, 2.52734375]])
        >>> best_iteration_results["current_line"]
        array([[1.85546875, 2.67396379, 2.08203125],
        [0.85546875, 2.57869077, 2.52734375]])
        >>> best_iteration_results["threshold"]
        0.02
        >>> best_iteration_results["number_inliers"]
        2128
        >>> best_iteration_results["indices_inliers"][:5]
        array([24335, 25743, 25746, 25897, 26405])
    '''
    if seed is not None:
        np.random.seed(seed)

    dict_full_results = {}
    dict_full_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    if audit_cloud:
        audit_before_sanitizing = pointcloud.pointcloud_audit(pcd)
        dict_full_results["audit_before_sanitizing"] = audit_before_sanitizing
        pcd = pointcloud.pointcloud_sanitize(pcd)
        audit_after_sanitizing = pointcloud.pointcloud_audit(pcd)
        dict_full_results["audit_after_sanitizing"] = audit_after_sanitizing
    
    number_pcd_points = len(pcd.points)
    dict_full_results["number_pcd_points"] = number_pcd_points
    np_points = np.asarray(pcd.points)

    dict_results_from_np_points = get_ransac_data_from_np_points(np_points, ransac_iterator = ransac_iterator, 
                                                                 ransac_iterations = ransac_iterations, 
                                                                 threshold = threshold, 
                                                                 verbosity_level = verbosity_level, 
                                                                 inherited_verbose_string = inherited_verbose_string,
                                                                 seed = seed)

    dict_full_results += dict_results_from_np_points
    return dict_full_results

# create a function that returns all the current_line along with their number_inliers

def get_lines_and_number_inliers_from_ransac_data_from_file(ransac_data_from_file):
    pair_lines_number_inliers = []
    for dict_iteration_results in ransac_data_from_file["ransac_iterations_results"]:
        pair_lines_number_inliers.append((dict_iteration_results["current_line"], dict_iteration_results["number_inliers"]))
    return pair_lines_number_inliers

# order the pairs by number_inliers

def get_lines_and_number_inliers_ordered_by_number_inliers(ransac_data: Dict) -> List[Tuple[np.ndarray, int]]:
    '''
    Gets the lines and number of inliers ordered by number of inliers from ransac data previously extracted.

    :param ransac_data: The ransac data.
    :type ransac_data: Dict
    :return: The list of pairs of lines and number of inliers.
    :rtype: List[Tuple[np.ndarray, int]]

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterator = coreransac.get_ransac_line_iteration_results
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> ransac_data = ransaclp.get_ransac_data_from_filename(office_filename, ransac_iterator = ransac_iterator,
                                                                 ransac_iterations = ransac_iterations,
                                                                 threshold = threshold, audit_cloud=True, seed = seed)
        >>> pair_lines_number_inliers_ordered = ransaclp.get_lines_and_number_inliers_ordered_by_number_inliers(ransac_data)
        >>> pair_lines_number_inliers_ordered[:5]
        [(array([[1.85546875, 2.67396379, 2.08203125],
                [0.85546875, 2.57869077, 2.52734375]]),
        2128),
        (array([[0.19140625, 2.10900044, 2.62890625],
                [1.03637648, 1.11328125, 2.43359375]]),
        1988),
        (array([[3.12890625, 2.17578125, 2.52116251],
                [1.27803993, 0.67578125, 2.33203125]]),
        1960),
        (array([[0.31464097, 0.90234375, 2.46484375],
                [2.99609375, 1.31640625, 2.34281754]]),
        1953),
        (array([[1.06640625, 0.86328125, 2.37622762],
                [1.03515625, 1.08984375, 2.42773271]]),
        1925)]


    '''
    pair_lines_number_inliers = get_lines_and_number_inliers_from_ransac_data_from_file(ransac_data)
    pair_lines_number_inliers_ordered = sorted(pair_lines_number_inliers, key=lambda pair_line_number_inliers: pair_line_number_inliers[1], reverse=True)
    return pair_lines_number_inliers_ordered

def get_ordered_list_sse_plane(pair_lines_number_inliers:List[Tuple[np.ndarray, int]], number_best: int = None, percentage_best: float = 0.2, 
                               already_ordered: bool = False, verbosity_level: int = 0, inherited_verbose_string: str = "") -> List[Tuple[float, np.ndarray]]:
    '''
    Gets a list of the sse (sum of squared errors) of the planes built from the best pairs of lines, along with the planes. 
    The best pairs of lines are the ones with the highest number of inliers. 
    The number of planes returned is either the number_best or the percentage_best of the total number of planes.

    :param pair_lines_number_inliers: The list of pairs of lines and number of inliers.
    :type pair_lines_number_inliers: List[Tuple[np.ndarray, int]]
    :param number_best: The number of best planes to consider.
    :type number_best: int
    :param percentage_best: The percentage of best planes to consider.
    :type percentage_best: float
    :param already_ordered: Whether the pairs are already ordered by number of inliers.
    :type already_ordered: bool
    :param verbosity_level: The verbosity level.
    :type verbosity_level: int
    :param inherited_verbose_string: The inherited verbose string.
    :type inherited_verbose_string: str
    :return: The list of sse and planes ordered by sse.
    :rtype: List[Tuple[float, np.ndarray]]

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterator = coreransac.get_ransac_line_iteration_results
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> ransac_data = ransaclp.get_ransac_data_from_filename(office_filename, ransac_iterator = ransac_iterator,
                                                                    ransac_iterations = ransac_iterations,
                                                                    threshold = threshold, audit_cloud=True, seed = seed)
        >>> pair_lines_number_inliers = ransaclp.get_lines_and_number_inliers_from_ransac_data_from_file(ransac_data)
        >>> list_sse_plane = ransaclp.get_ordered_list_sse_plane(pair_lines_number_inliers, percentage_best = 0.2)
        >>> list_sse_plane[:5]
        [(3.133004699296308e-09,
        array([ 0.02115254,  0.23488194, -0.97179373,  2.07212126])),
        (4.072922344774246e-09,
        array([-0.0583794 ,  0.16947913, -0.98380317,  2.274669  ])),
        (5.597729648076989e-09,
        array([ 0.04009822, -0.15939724,  0.98639984, -2.26464593])),
        (1.6036620702538505e-08,
        array([ 0.04939934, -0.14154284,  0.98869881, -2.30467834])),
        (1.7597194644519755e-08,
        array([-0.02745968,  0.98751003,  0.15514477, -2.9031455 ]))]
    '''
    if number_best is None:
        number_best = int(len(pair_lines_number_inliers) * percentage_best)
    if not already_ordered:
        pair_lines_number_inliers = sorted(pair_lines_number_inliers, key=lambda x: x[1], reverse=True)
    list_sse_plane = []
    current_iteration = 0
    # compute the maximum number of iterations
    total_number_iterations = int(number_best * (number_best - 1) / 2)
    for i in range(number_best):
        for j in range(i+1, number_best):
            # compute the current number of iterations
            current_iteration += 1
            message = f"Estimating planes: Line pair {current_iteration}/{total_number_iterations}"
            if verbosity_level == 1 and current_iteration % 10 == 0 or verbosity_level >= 2:
                message = f"{inherited_verbose_string} {message}"
                print(message)
            line_1 = pair_lines_number_inliers[i][0]
            line_2 = pair_lines_number_inliers[j][0]
            plane, error = geom.get_best_plane_from_points_from_two_segments(line_1, line_2)
            list_sse_plane.append((error, plane))
    # order the list by sse
    list_sse_plane = sorted(list_sse_plane, key=lambda x: x[0])
    return list_sse_plane

def get_n_percentile_from_list_sse_plane(list_sse_plane: List[Tuple[float, np.ndarray]], percentile: int = 5) -> List[Tuple[float, np.ndarray]]:
    '''
    Gets the list of elements in which sse is below the n percentile from a list of sse and planes.
    The function looks for the best sse values, that are the lowest ones.

    :param list_sse_plane: The list of sse and planes.
    :type list_sse_plane: List[Tuple[float, np.ndarray]]
    :param percentile: The percentile.
    :type percentile: int
    :return: The n percentile.
    :rtype: List[Tuple[float, np.ndarray]]

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterator = coreransac.get_ransac_line_iteration_results
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> ransac_data = ransaclp.get_ransac_data_from_filename(office_filename, ransac_iterator = ransac_iterator,
                                                                    ransac_iterations = ransac_iterations,
                                                                    threshold = threshold, audit_cloud=True, seed = seed)
        >>> pair_lines_number_inliers = ransaclp.get_lines_and_number_inliers_from_ransac_data_from_file(ransac_data)
        >>> list_sse_plane = ransaclp.get_ordered_list_sse_plane(pair_lines_number_inliers, percentage_best = 0.2)
        >>> len(list_sse_plane)
        780
        >>> list_sse_plane[:5]
        [(3.133004699296308e-09,
        array([ 0.02115254,  0.23488194, -0.97179373,  2.07212126])),
        (4.072922344774246e-09,
        array([-0.0583794 ,  0.16947913, -0.98380317,  2.274669  ])),
        (5.597729648076989e-09,
        array([ 0.04009822, -0.15939724,  0.98639984, -2.26464593])),
        (1.6036620702538505e-08,
        array([ 0.04939934, -0.14154284,  0.98869881, -2.30467834])),
        (1.7597194644519755e-08,
        array([-0.02745968,  0.98751003,  0.15514477, -2.9031455 ]))]
        >>> list_sse_plane_05 = ransaclp.get_n_percentile_from_list_sse_plane(list_sse_plane, percentile = 5)
        >>> len(list_sse_plane_05)
        39
        >>> list_sse_plane_05[:5]
        [(3.133004699296308e-09,
        array([ 0.02115254,  0.23488194, -0.97179373,  2.07212126])),
        (4.072922344774246e-09,
        array([-0.0583794 ,  0.16947913, -0.98380317,  2.274669  ])),
        (5.597729648076989e-09,
        array([ 0.04009822, -0.15939724,  0.98639984, -2.26464593])),
        (1.6036620702538505e-08,
        array([ 0.04939934, -0.14154284,  0.98869881, -2.30467834])),
        (1.7597194644519755e-08,
        array([-0.02745968,  0.98751003,  0.15514477, -2.9031455 ]))]

    '''
    list_sse, list_plane = zip(*list_sse_plane)  # Unpack the list of tuples
    percentile_threshold = np.percentile(list_sse, percentile)
    filtered_pairs = [(sse, plane) for sse, plane in zip(list_sse, list_plane) if sse <= percentile_threshold]
    return filtered_pairs

def get_ransaclp_data_from_filename(filename: str, ransac_iterations: int = 100, threshold: float = 0.1, audit_cloud: bool = False, 
                                    verbosity_level: int = 0, inherited_verbose_string: str = "", seed: int = None) -> Dict:
    '''
    Gets the ransaclp data from a file.
    
    :param filename: The filename.
    :type filename: str
    :param ransac_iterations: The number of ransac iterations.
    :type ransac_iterations: int
    :param threshold: The threshold.
    :type threshold: float
    :param audit_cloud: Whether to audit the cloud.
    :type audit_cloud: bool
    :param verbosity_level: The verbosity level.
    :type verbosity_level: int
    :param inherited_verbose_string: The inherited verbose string.
    :type inherited_verbose_string: str
    :param seed: The seed.
    :type seed: int
    :return: The ransaclp data.
    :rtype: dict

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> ransaclp_data = ransaclp.get_ransaclp_data_from_filename(office_filename, 
                                                                    ransac_iterations = ransac_iterations, 
                                                                    threshold = threshold, audit_cloud=True, seed = seed)
        >>> ransaclp_data
        {'plane': array([-0.07208637,  0.20357587, -0.97640177,  2.25114528]),
        'number_inliers': 83788,
        'indices_inliers': array([     7,      8,     10, ..., 248476, 248477, 248478])}
        >>> pcd = o3d.io.read_point_cloud(office_filename)
        >>> number_inliers = ransaclp_data["number_inliers"]
        >>> indices_inliers = ransaclp_data["indices_inliers"]
        >>> inlier_cloud = pcd.select_by_index(indices_inliers)
        >>> inlier_cloud.paint_uniform_color([1, 0, 0])
        >>> o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="RANSACLP Inliers:  " + str(number_inliers))

    |ransaclp_get_ransaclp_data_from_filename_example|

    .. |ransaclp_get_ransaclp_data_from_filename_example| image:: ../../_static/images/ransaclp_get_ransaclp_data_from_filename_example.png

    '''
    if seed is not None:
        np.random.seed(seed)
    ransac_iterator = coreransac.get_ransac_line_iteration_results
    ransac_data = get_ransac_data_from_filename(filename, ransac_iterator = ransac_iterator,
                                                ransac_iterations = ransac_iterations,
                                                threshold = threshold, audit_cloud=audit_cloud, verbosity_level = verbosity_level, 
                                                inherited_verbose_string=inherited_verbose_string, seed = seed)
    pcd = o3d.io.read_point_cloud(ransac_data["filename"])
    np_points = np.asarray(pcd.points)

    pair_lines_number_inliers = get_lines_and_number_inliers_from_ransac_data_from_file(ransac_data)
    ordered_list_sse_plane = get_ordered_list_sse_plane(pair_lines_number_inliers, percentage_best = 0.2, verbosity_level=verbosity_level,
                                                        inherited_verbose_string=inherited_verbose_string)
    list_sse_plane_05 = get_n_percentile_from_list_sse_plane(ordered_list_sse_plane, percentile = 5)
    list_good_planes = [sse_plane[1] for sse_plane in list_sse_plane_05]
    results_from_best_plane = coreransac.get_best_fitting_data_from_list_planes(np_points, list_good_planes, threshold)
    return results_from_best_plane

def get_ransaclp_data_from_np_points(np_points: np.ndarray, ransac_iterations: int = 100, threshold: float = 0.1,
                                     cuda: bool = False, 
                                    verbosity_level: int = 0, inherited_verbose_string: str = "", seed: int = None) -> Dict:
    '''
    Gets the ransaclp data from a file.
    
    :param np_points: The numpy array of points.
    :type np_points: np.ndarray
    :param ransac_iterations: The number of ransac iterations.
    :type ransac_iterations: int
    :param threshold: The threshold.
    :type threshold: float
    :param cuda: Wheter cuda is used.
    :type cuda: bool
    :param verbosity_level: The verbosity level.
    :type verbosity_level: int
    :param inherited_verbose_string: The inherited verbose string.
    :type inherited_verbose_string: str
    :param seed: The seed.
    :type seed: int
    :return: The ransaclp data.
    :rtype: dict

    :Example:

    ::

        >>> import open3d as o3d
        >>> import mrdja.ransaclp as ransaclp
        >>> import mrdja.ransac.coreransac as coreransac
        >>> office_dataset = o3d.data.OfficePointClouds()
        >>> office_filename = office_dataset.paths[0]
        >>> ransac_iterations = 200
        >>> threshold = 0.02
        >>> seed = 42
        >>> pcd = o3d.io.read_point_cloud(office_filename)
        >>> np_points = np.asarray(pcd.points)
        >>> ransaclp_data = ransaclp.get_ransaclp_data_from_filename(np_points, 
                                                                    ransac_iterations = ransac_iterations, 
                                                                    threshold = threshold, audit_cloud=True, seed = seed)
        >>> ransaclp_data
        {'plane': array([-0.07208637,  0.20357587, -0.97640177,  2.25114528]),
        'number_inliers': 83788,
        'indices_inliers': array([     7,      8,     10, ..., 248476, 248477, 248478])}
        >>> pcd = o3d.io.read_point_cloud(office_filename)
        >>> number_inliers = ransaclp_data["number_inliers"]
        >>> indices_inliers = ransaclp_data["indices_inliers"]
        >>> inlier_cloud = pcd.select_by_index(indices_inliers)
        >>> inlier_cloud.paint_uniform_color([1, 0, 0])
        >>> o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="RANSACLP Inliers:  " + str(number_inliers))

    |ransaclp_get_ransaclp_data_from_filename_example|

    .. |ransaclp_get_ransaclp_data_from_filename_example| image:: ../../_static/images/ransaclp_get_ransaclp_data_from_filename_example.png

    '''
    if seed is not None:
        np.random.seed(seed)
    if cuda:
        ransac_iterator = coreransaccuda.get_ransac_line_iteration_results_cuda
        ransac_data = get_ransac_data_from_np_points_cuda(np_points, ransac_iterator = ransac_iterator,
                                            ransac_iterations = ransac_iterations,
                                            threshold = threshold,
                                            verbosity_level = verbosity_level, 
                                            inherited_verbose_string=inherited_verbose_string, seed = seed)
    else:
        ransac_iterator = coreransac.get_ransac_line_iteration_results
        ransac_data = get_ransac_data_from_np_points(np_points, ransac_iterator = ransac_iterator,
                                                    ransac_iterations = ransac_iterations,
                                                    threshold = threshold,
                                                    verbosity_level = verbosity_level, 
                                                    inherited_verbose_string=inherited_verbose_string, seed = seed)

    pair_lines_number_inliers = get_lines_and_number_inliers_from_ransac_data_from_file(ransac_data)
    ordered_list_sse_plane = get_ordered_list_sse_plane(pair_lines_number_inliers, percentage_best = 0.2, verbosity_level=verbosity_level,
                                                        inherited_verbose_string=inherited_verbose_string)
    list_sse_plane_05 = get_n_percentile_from_list_sse_plane(ordered_list_sse_plane, percentile = 5)
    list_good_planes = [sse_plane[1] for sse_plane in list_sse_plane_05]
    results_from_best_plane = coreransac.get_best_fitting_data_from_list_planes(np_points, list_good_planes, threshold)
    return results_from_best_plane


'''
**RANSAC** core functions for finding geometric primitives (**planes** and **lines**) in point clouds.

In this module all the functions related to the **RANSAC** algorithm for **planes** and **lines** are defined.

The **pointcloud** is expected to be a **numpy array** of shape **(N, 3)** where **N** is the number of **points** and **3** is the 
number of **coordinates (x, y, z)**. 
**No** information about **RGB** values or normals are taken into account. Therefore, an **Open3D** point cloud must be **converted**
to a **numpy array** before 
using the functions in this module.

The following code is an example of how to extract the points from an Open3D point cloud and convert them to a numpy array:

::

    import numpy as np
    import open3d as o3d

    filename = "pointcloud.ply"
    pcd = o3d.io.read_point_cloud(filename)
    np_points = np.asarray(pcd.points)

The pointcloud could be in **ply* or **pcd** format, because Open3D can read both formats.

You can always access to the pointclouds provided by Open3D. For example, the following code shows how to access to the LivingRoomPointClouds, 
which are 57 point clouds of from the Redwood RGB-D Dataset.

::

    import open3d as o3d
    dataset = o3d.data.LivingRoomPointClouds()
    pcds_living_rooms = []
    for pcd_path in dataset.paths:
        pcds_living_rooms.append(o3d.io.read_point_cloud(pcd_path))

Other 53 pointclouds from the same dataset are also available in OfficePointClouds:

::

    import open3d as o3d
    dataset = o3d.data.OfficePointClouds()
    pcds_offices = []
    for pcd_path in dataset.paths:
        pcds_offices.append(o3d.io.read_point_cloud(pcd_path))

After executing the previous code, we could define two examples of a living room pointcloud and an office pointcloud, respectively:

::

    living_room_pcd = pcds_living_rooms[0]
    office_pcd = pcds_offices[0]

    
And visualize them:

::

    o3d.visualization.draw_geometries([living_room_pcd])

    
|coreransac_living_room_pcd|

    .. |coreransac_living_room_pcd| image:: ../../_static/images/coreransac_living_room_pcd.png

::

    o3d.visualization.draw_geometries([office_pcd])

|coreransac_office_pcd|

    .. |coreransac_office_pcd| image:: ../../_static/images/coreransac_office_pcd.png


A brief explanation of the relationship between the functions defined here is given below:

If we want to extract the best fitting plane from a pointcloud, we have to call the function :func:`get_ransac_plane_results`, which, given
a pointcloud, the maximum distance from a point to the plane for it to be considered an inlier, and the number of iterations to run the RANSAC algorithm,
returns the best plane parameters, the number of inliers, and their indices. It calls the function :func:`get_ransac_plane_iteration_results` to get the results
of each iteration of the RANSAC algorithm. 

'''

import numpy as np
import open3d as o3d
import math
import mrdja.ransac.coreransacutils as crsu
import mrdja.sampling as sampling
from typing import Optional, List, Tuple, Dict, Union
import mrdja.geometry as geom

def get_how_many_below_threshold_between_line_and_points_and_their_indices(points: np.ndarray, line_two_points: np.ndarray, threshold: np.float32) -> Tuple[int, np.ndarray]:
    """
    Computes how many points are below a threshold distance from a line and returns their count and their indices.

    :param points: The collection of points to measure the distance to the line.
    :type points: np.ndarray
    :param line_two_points: Two points defining the line.
    :type line_two_points: np.ndarray
    :param threshold: Maximum distance to the line.
    :type threshold: np.float32
    :return: Number of points below the threshold distance as their indices.
    :rtype: Tuple[int, np.ndarray]

    :Example:

    ::

        >>> import mrdja.ransac.coreransac as coreransac
        >>> import mrdja.geometry as geom
        >>> import mrdja.drawing as drawing
        >>> import mrdja.matplot3d as plot3d
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> line = np.array([[0, 0, 0], [1, 1, 1]])
        >>> points = np.array([[-1, -1, -1], [0, 0, 0], [1, 1, 1], [2, 2, 2], [2, 2, 3], [2, 2, 4]])
        >>> threshold = 1
        >>> count, indices_below = coreransac.get_how_many_below_threshold_between_line_and_points_and_their_indices(points, line, threshold)
        >>> count
        5
        >>> indices_below
        array([0, 1, 2, 3, 4])
        >>> points[indices_below]
        array([[-1, -1, -1],
        [ 0,  0,  0],
        [ 1,  1,  1],
        [ 2,  2,  2],
        [ 2,  2,  3]])

        >>> # Calculate the minimum and maximum points of the cube.
        >>> cube_min = np.min(points, axis=0)
        >>> cube_max = np.max(points, axis=0)

        >>> # Get the intersection of the line with the cube
        >>> intersection_points = geom.get_intersection_points_of_line_with_cube(line, cube_min, cube_max)

        >>> # Draw the cube
        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111, projection='3d')
        >>> drawing.draw_cube(cube_min, cube_max, alpha = 0.1, ax = ax)
        >>> # Draw the segment between the intersection points
        >>> plot3d.draw_segment(intersection_points, color="green", ax = ax)
        >>> # Draw the intersection points
        >>> plot3d.draw_points(intersection_points, color="black", style="X", ax = ax)


        >>> # Get the indices of the points above the threshold
        >>> all_indices = np.arange(len(points))
        >>> indices_above = np.setdiff1d(all_indices, indices_below)
        >>> # Plot the points below the threshold in red
        >>> plot3d.draw_points(points[indices_below], color="red", style="o", ax = ax)
        >>> # Plot the points above the threshold in blue
        >>> plot3d.draw_points(points[indices_above], color="blue", style="o", ax = ax)

        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Y')
        >>> ax.set_zlabel('Z')
        >>> ax.legend()
        >>> plt.show()

    |coreransac_get_how_many_below_threshold_between_line_and_points_and_their_indices_example|

    .. |coreransac_get_how_many_below_threshold_between_line_and_points_and_their_indices_example| image:: ../../_static/images/coreransac_get_how_many_below_threshold_between_line_and_points_and_their_indices_example.png


    """
    B = line_two_points[0]
    C = line_two_points[1]
    # the distance between a point A and the line defined by the points B and C can be computed as 
    # magnitude(cross(A - B, C - B)) / magnitude(C - B)
    # https://math.stackexchange.com/questions/1905533/find-perpendicular-distance-from-point-to-line-in-3d

    cross_product = np.cross(points - B, C - B)
    magnitude_cross_product = np.linalg.norm(cross_product, axis=1)
    magnitude_C_minus_B = np.linalg.norm(C - B)
    distance = magnitude_cross_product / magnitude_C_minus_B

    indices_inliers = np.array([index for index, value in enumerate(distance) if value <= threshold], dtype=np.int64)
    return len(indices_inliers), indices_inliers

def get_how_many_below_threshold_between_plane_and_points_and_their_indices(points: np.ndarray, plane: np.ndarray, threshold: np.float32) -> Tuple[int, np.ndarray]:
    """
    Computes how many points are below a threshold distance from a plane and returns their count and their indices.

    :param points: The collection of points to measure the distance to the line.
    :type points: np.ndarray
    :param plane: Four parameters defining the plane in the form Ax + By + Cz + D = 0.
    :type plane: np.ndarray
    :param threshold: Maximum distance to the line.
    :type threshold: np.float32
    :return: Number of points below the threshold distance as their indices.
    :rtype: Tuple[int, np.ndarray]

    :Example:

    ::

        >>> import mrdja.ransac.coreransac as coreransac
        >>> import numpy as np
        >>> plane = np.array([1, 1, 1, 0])
        >>> points = np.array([[-1, -1, -1], [0, 0, 0], [1, 1, 1], [2, 2, 2], [2, 2, 3], [2, 2, 4]])
        >>> threshold = 1
        >>> count, indices = coreransac.get_how_many_below_threshold_between_plane_and_points_and_their_indices(points, plane, threshold)
        >>> count
        1
        >>> indices
        array([1])
        >>> points[indices]
        array([[0, 0, 0]])
    """
    a = plane[0]
    b = plane[1]
    c = plane[2]
    d = plane[3]
    points_x = points[:, 0]
    points_y = points[:, 1]
    points_z = points[:, 2]
    # the distance between the point P with coordinates (xo, yo, zo) and the given plane with equation Ax + By + Cz = D, 
    # is given by |Axo + Byo+ Czo + D|/√(A^2 + B^2 + C^2).
    denominator = math.sqrt(a * a + b * b + c * c)
    optimized_threshold = threshold * denominator  # for not computing the division for each point
    distance = np.abs(points_x * a + points_y * b + points_z * c + d)
    # point_indices = np.where(distance <= optimized_threshold)[0]
    indices_inliers = np.array([index for index, value in enumerate(distance) if value <= optimized_threshold], dtype=np.int64)
    return len(indices_inliers), indices_inliers

def get_ransac_line_iteration_results(points: np.ndarray, threshold: float, len_points:Optional[int]=None, seed: Optional[int]=None) -> dict:
    """
    Returns the results of one iteration of the RANSAC algorithm for line fitting.
    
    :param points: The collection of points to fit the line to.
    :type points: np.ndarray
    :param threshold: The maximum distance from a point to the line for it to be considered an inlier.
    :type threshold: float
    :param len_points: The number of points in the collection of points.
    :type len_points: Optional[int]
    :return: A dictionary containing the current line parameters, number of inliers, and their indices.
    :rtype: dict
    """
    if len_points is None:
        len_points = len(points)
    if seed is not None:
        np.random.seed(seed)
    current_random_points = sampling.sampling_np_arrays_from_enumerable(points, cardinality_of_np_arrays=2, number_of_np_arrays=1, num_source_elems=len_points, seed=seed)[0]
    current_line = current_random_points # the easiest way to get the line parameters
    how_many_in_line, current_point_indices = get_how_many_below_threshold_between_line_and_points_and_their_indices(points, current_line, threshold)
    # print(len_points, current_random_points, current_plane, how_many_in_plane)
    return {"current_random_points": current_random_points, "current_line": current_line, "threshold": threshold, "number_inliers": how_many_in_line, "indices_inliers": current_point_indices}


def get_ransac_plane_iteration_results(points: np.ndarray, threshold: float, len_points:Optional[int]=None, seed: Optional[int]=None) -> dict:
    """
    Returns the results of one iteration of the RANSAC algorithm for plane fitting.

    This functions expects a **collection of points**, the **number of points** in the collection, the **maximum distance** from a point
    to the plane for it to be considered an inlier, and the **seed** to initialize the random number generator.
    It returns a dictionary containing the **current plane parameters**, **number of inliers**, and **their indices**. The keys of the dictionary
    are **current_random_points**, **"current_plane"**, **"threshold"**, **"number_inliers"**, and **"indices_inliers"**, respectively. 
    The type of the values of the dictionary are **np.ndarray**, **np.ndarray**, **float**, **int**, and **np.ndarray**, respectively.
    
    :param points: The collection of points to fit the plane to.
    :type points: np.ndarray
    :param threshold: The maximum distance from a point to the plane for it to be considered an inlier.
    :type threshold: float
    :param len_points: The number of points in the collection of points.
    :type len_points: Optional[int]
    :param seed: The seed to initialize the random number generator.
    :type seed: Optional[int]
    :return: A dictionary containing the current plane parameters, number of inliers, and their indices, as well as the three random points sampled to create the plane.
    :rtype: dict

    :Example:

    ::

        >>> import mrdja.ransac.coreransac as coreransac
        >>> import open3d as o3d
        >>> import numpy as np
        >>> import random
        >>> import open3d as o3d
        >>> dataset = o3d.data.OfficePointClouds()
        >>> pcds_offices = []
        >>> for pcd_path in dataset.paths:
        >>>     pcds_offices.append(o3d.io.read_point_cloud(pcd_path))
        >>> office_pcd = pcds_offices[0]
        >>> pcd_points = np.asarray(office_pcd.points)
        >>> threshold = 0.1
        >>> num_iterations = 20
        >>> dict_results = coreransac.get_ransac_plane_iteration_results(pcd_points, threshold, seed = 42)
        >>> dict_results
        >>> {'current_random_points': array([[1.61072648, 1.83984375, 1.91796875],
        >>> [3.00390625, 2.68674755, 2.01953125],
        >>> [2.10068583, 2.34765625, 2.14453125]]),
        >>> 'current_plane': array([ 0.14030194, -0.2658808 ,  0.29252566, -0.297864  ]),
        >>> 'threshold': 0.1,
        >>> 'number_inliers': 34283,
        >>> 'indices_inliers': array([ 98356, 101924, 101956, ..., 271055, 271245, 271246])}
        >>> inliers = dict_results["indices_inliers"]
        >>> inlier_cloud = office_pcd.select_by_index(inliers)
        >>> inlier_cloud.paint_uniform_color([1.0, 0, 0])
        >>> outlier_cloud = office_pcd.select_by_index(inliers, invert=True)
        >>> o3d.visualization.draw_geometries([inlier_cloud, outlier_cloud])

    |coreransac_get_ransac_plane_iteration_results_example|

    .. |coreransac_get_ransac_plane_iteration_results_example| image:: ../../_static/images/coreransac_get_ransac_plane_iteration_results_example.png

    """
    if len_points is None:
        len_points = len(points)
    if seed is not None:
        np.random.seed(seed)
    current_random_points = sampling.sampling_np_arrays_from_enumerable(points, cardinality_of_np_arrays=3, number_of_np_arrays=1, num_source_elems=len_points, seed=seed)[0]
    current_plane = geom.get_plane_from_list_of_three_points(current_random_points.tolist())
    how_many_in_plane, current_point_indices = get_how_many_below_threshold_between_plane_and_points_and_their_indices(points, current_plane, threshold)
    # print(len_points, current_random_points, current_plane, how_many_in_plane)
    return {"current_random_points": current_random_points, "current_plane": current_plane, "threshold": threshold, "number_inliers": how_many_in_plane, "indices_inliers": current_point_indices}
    

def get_ransac_plane_results(points: np.ndarray, threshold: float, num_iterations: int, len_points:Optional[int]=None, seed: Optional[int] = None) -> dict:
    """
    Computes the **best plane** that fits a **collection of points** and the indices of the inliers.

    This functions expects a **collection of points**, the **number of points** in the collection, the **maximum distance** from a point 
    to the plane for it to be considered an inlier, and the **number of iterations** to run the RANSAC algorithm. 
    It returns a **dictionary** containing the
    **best plane parameters**, **number of inliers**, and **their indices**. The keys of the dictionary are **"best_plane"**, 
    **"number_inliers"**, and
    **"indices_inliers"**, respectively. The type of the values of the dictionary are **np.ndarray**, **int**, and **np.ndarray**, respectively. 
    Reproducibility is enforced by setting the **seed** parameter to a fixed value.

    :param points: The collection of points to fit the plane to.
    :type points: np.ndarray
    :param threshold: The maximum distance from a point to the plane for it to be considered an inlier.
    :type threshold: float
    :param num_iterations: The number of iterations to run the RANSAC algorithm.
    :type num_iterations: int
    :param len_points: The number of points in the collection of points.
    :type len_points: Optional[int]
    :param seed: The seed to initialize the random number generator.
    :type seed: Optional[int]
    :return: A dictionary containing the best plane parameters, number of inliers, and their indices.
    :rtype: dict

    :Example:

    ::

        >>> import mrdja.ransac.coreransac as coreransac
        >>> import open3d as o3d
        >>> import numpy as np
        >>> import random
        >>> import open3d as o3d
        >>> dataset = o3d.data.OfficePointClouds()
        >>> pcds_offices = []
        >>> for pcd_path in dataset.paths:
        >>>     pcds_offices.append(o3d.io.read_point_cloud(pcd_path))
        >>> office_pcd = pcds_offices[0]
        >>> pcd_points = np.asarray(office_pcd.points)
        >>> threshold = 0.1
        >>> num_iterations = 20
        >>> dict_results = coreransac.get_ransac_plane_results(pcd_points, threshold, num_iterations, seed = 42)
        >>> dict_results
        {'best_plane': array([-0.17535096,  0.45186984, -2.44615646,  5.69205427]),
        'number_inliers': 153798,
        'indices_inliers': array([     0,      1,      2, ..., 248476, 248477, 248478])}
        >>> inliers = dict_results["indices_inliers"]
        >>> inlier_cloud = office_pcd.select_by_index(inliers)
        >>> inlier_cloud.paint_uniform_color([1.0, 0, 0])
        >>> outlier_cloud = office_pcd.select_by_index(inliers, invert=True)
        >>> o3d.visualization.draw_geometries([inlier_cloud, outlier_cloud])

    |coreransac_get_ransac_plane_results_example|

    .. |coreransac_get_ransac_plane_results_example| image:: ../../_static/images/coreransac_get_ransac_plane_results_example.png


    """
    if len_points is None:
        len_points = len(points)
    if seed is not None:
        np.random.seed(seed)
    best_plane = None
    number_points_in_best_plane = 0
    for _ in range(num_iterations):
        dict_results = get_ransac_plane_iteration_results(points, threshold, len_points)
        current_plane = dict_results["current_plane"]
        how_many_in_plane = dict_results["number_inliers"]
        current_indices_inliers = dict_results["indices_inliers"]
        if how_many_in_plane > number_points_in_best_plane:
            inliers_ratio = how_many_in_plane / len_points
            max_num_iterations = crsu.compute_number_iterations(inliers_ratio, alpha = 0.05)
            print("Current inliers ratio: ", inliers_ratio, " Max num iterations: ", max_num_iterations)
            number_points_in_best_plane = how_many_in_plane
            best_plane = current_plane
            indices_inliers = current_indices_inliers.copy()
    return {"best_plane": best_plane, "number_inliers": number_points_in_best_plane, "indices_inliers": indices_inliers}


def get_fitting_data_from_list_planes(points: np.ndarray, list_planes: List[np.ndarray], threshold: float) -> List[Dict]:
    '''
    Returns the fitting data for each plane in the list of planes.

    :param points: The collection of points to fit the plane to.
    :type points: np.ndarray
    :param list_planes: The list of planes to fit to the points.
    :type list_planes: List[np.ndarray]
    :param threshold: The maximum distance from a point to the plane for it to be considered an inlier.
    :type threshold: float
    :return: A list of dictionaries containing the plane parameters, number of inliers, and their indices.
    :rtype: List[Dict]

    :Example:

    ::

        >>> import mrdja.ransac.coreransac as coreransac
        >>> import open3d as o3d
        >>> import numpy as np
        >>> import random
        >>> import open3d as o3d
        >>> dataset = o3d.data.OfficePointClouds()
        >>> pcds_offices = []
        >>> for pcd_path in dataset.paths:
        >>>     pcds_offices.append(o3d.io.read_point_cloud(pcd_path))
        >>> office_pcd = pcds_offices[0]
        >>> pcd_points = np.asarray(office_pcd.points)
        >>> threshold = 0.1
        >>> num_iterations = 20
        >>> dict_results = coreransac.get_ransac_plane_results(pcd_points, threshold, num_iterations, seed = 42)
        >>> dict_results
        {'best_plane': array([-0.17535096,  0.45186984, -2.44615646,  5.69205427]),
        'number_inliers': 153798,
        'indices_inliers': array([     0,      1,      2, ..., 248476, 248477, 248478])}
        >>> fitting_data = coreransac.get_fitting_data_from_list_planes(pcd_points, [dict_results["best_plane"]], threshold)
        >>> fitting_data
        [{'plane': array([-0.17535096,  0.45186984, -2.44615646,  5.69205427]),
        'number_inliers': 153798,
        'indices_inliers': array([     0,      1,      2, ..., 248476, 248477, 248478])}]
    ''' 
    list_fitting_data = []
    for plane in list_planes:
        how_many_in_plane, indices_inliers = get_how_many_below_threshold_between_plane_and_points_and_their_indices(points, plane, threshold)
        list_fitting_data.append({"plane": plane, "number_inliers": how_many_in_plane, "indices_inliers": indices_inliers})
    return list_fitting_data

def get_best_fitting_data_from_list_planes(points: np.ndarray, list_planes: List[np.ndarray], threshold: float) -> Dict:
    '''
    Returns the fitting data for the best plane in the list of planes.

    :param points: The collection of points to fit the plane to.
    :type points: np.ndarray
    :param list_planes: The list of planes to fit to the points.
    :type list_planes: List[np.ndarray]
    :param threshold: The maximum distance from a point to the plane for it to be considered an inlier.
    :type threshold: float
    :return: A dictionary containing the plane parameters, number of inliers, and their indices.
    :rtype: Dict

    :Example:

    ::

        >>> import mrdja.ransac.coreransac as coreransac
        >>> import open3d as o3d
        >>> import numpy as np
        >>> import random
        >>> import open3d as o3d
        >>> dataset = o3d.data.OfficePointClouds()
        >>> pcds_offices = []
        >>> for pcd_path in dataset.paths:
        >>>     pcds_offices.append(o3d.io.read_point_cloud(pcd_path))
        >>> office_pcd = pcds_offices[0]
        >>> pcd_points = np.asarray(office_pcd.points)
        >>> threshold = 0.1
        >>> num_iterations = 20
        >>> dict_results = coreransac.get_ransac_plane_results(pcd_points, threshold, num_iterations, seed = 42)
        >>> dict_results
        {'best_plane': array([-0.17535096,  0.45186984, -2.44615646,  5.69205427]),
        'number_inliers': 153798,
        'indices_inliers': array([     0,      1,      2, ..., 248476, 248477, 248478])}
        >>> fitting_data = coreransac.get_fitting_data_from_list_planes(pcd_points, [dict_results["best_plane"]], threshold)
        >>> fitting_data
        [{'plane': array([-0.17535096,  0.45186984, -2.44615646,  5.69205427]),
        'number_inliers': 153798,
        'indices_inliers': array([     0,      1,      2, ..., 248476, 248477, 248478])}]
        >>> best_fitting_data = coreransac.get_best_fitting_data_from_list_planes(pcd_points, [dict_results["best_plane"]], threshold)
        >>> best_fitting_data
        {'plane': array([-0.17535096,  0.45186984, -2.44615646,  5.69205427]),
        'number_inliers': 153798,
        'indices_inliers': array([     0,      1,      2, ..., 248476, 248477, 248478])}
    '''
    fitting_data = get_fitting_data_from_list_planes(points, list_planes, threshold)
    best_fitting_data = max(fitting_data, key=lambda fitting_data: fitting_data["number_inliers"])
    return best_fitting_data


import mrdja.ransac.coreransac as coreransac
import mrdja.ransaclp as ransaclp
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geom
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np
import random

'''
Line-guided plane detection in pointclouds.
Tomar, por ejemplo, 200 rectas y luego los 40 mejores por parejas. Eso son 780 parejas. Tomar el 5% mejor según el error de ajuste del plano.
Eso son 39 parejas. En total 239 veces que hay que mirar los inliers mirando todos los puntos.
Calcular el mejor plano, tomando RANSAC de open3d con 100.000 iteraciones.
Hacer varios experimentos, tanto de RANSAC con 239 o más iteraciones, como de open3d RANSAC con otros números de iteraciones.
'''

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

ransac_iterations = 200
threshold = 0.02
threshold_pcd = 0.02
seed = 42
RANSAC_iterator = coreransac.get_ransac_line_iteration_results

# filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
# filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/office_1/office_1.ply"
filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"
RANSAC_data = ransaclp.get_RANSAC_data_from_filename(filename, RANSAC_iterator = RANSAC_iterator, 
                                                           ransac_iterations = ransac_iterations, 
                                                           threshold = threshold, audit_cloud=False, seed = seed)

print(RANSAC_data)

pcd = o3d.io.read_point_cloud(RANSAC_data["filename"])
np_points = np.asarray(pcd.points)

ordered_pairs = ransaclp.get_lines_and_number_inliers_ordered_by_number_inliers(RANSAC_data)
pair_lines_number_inliers = ransaclp.get_lines_and_number_inliers_from_RANSAC_data_from_file(RANSAC_data)
ordered_list_sse_plane = ransaclp.get_ordered_list_sse_plane(pair_lines_number_inliers, percentage_best = 0.2)
list_sse_plane_05 = ransaclp.get_n_percentile_from_list_sse_plane(ordered_list_sse_plane, percentile = 5)
list_good_planes = [sse_plane[1] for sse_plane in list_sse_plane_05]
results_from_best_plane = coreransac.get_best_fitting_data_from_list_planes(np_points, list_good_planes, threshold_pcd)

number_inliers = results_from_best_plane["number_inliers"]
indices_inliers = results_from_best_plane["indices_inliers"]
inlier_cloud = pcd.select_by_index(indices_inliers)
inlier_cloud.paint_uniform_color([1, 0, 0])
o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers:  " + str(number_inliers))

# set random seed
np.random.seed(42)
plane_model, inliers = pcd.segment_plane(distance_threshold=0.02,
                                         ransac_n=3,
                                         num_iterations=478)
[a, b, c, d] = plane_model
print("Plane equation: {:.2f}x + {:.2f}y + {:.2f}z + {:.2f} = 0".format(a, b, c, d))
inlier_cloud = pcd.select_by_index(inliers)
inlier_cloud.paint_uniform_color([1, 0, 0])
o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers:  " + str(len(inliers)))
print("Number of inliers: " + str(len(inliers)))
print("Number of inliers maximum: " + str(number_inliers))






import os
import pickle
import pandas as pd
import numpy as np
import glob
from mrdja.stats import friedman_nemenyi_test

# Define the directory where the data files are located
data_dir = "/home/scpmaotj/pkl_files/me/"  # replace with your directory

# Initialize an empty dictionary to hold all the data
all_data = {}

# Iterate over all files in the directory that match the pattern '*means*.pkl'
for file_path in glob.glob(os.path.join(data_dir, '*percentage*.pkl')):
    # Load data from the current file
    with open(file_path, 'rb') as f:
        data = pickle.load(f)
    
    # Add the loaded data to the all_data dictionary
    all_data[os.path.basename(file_path)] = data

# Convert the dictionary of dictionaries into a DataFrame
df = pd.DataFrame(all_data).T

results = friedman_nemenyi_test(df)

nemenyi_df = results["nemenyi_df"]
nemenyi_table = results["nemenyi_table"]
friedman_result = results["friedman_result"]
mean_ranks = results["mean_ranks"]
nemenyi_result = results["nemenyi_result"]
nemenyi_df_2 = results["nemenyi_df_2"]
nemenyi_table_2 = results["nemenyi_table_2"]

print("Friedman result:", friedman_result)
print("Mean ranks:\n", mean_ranks)
print("Nemenyi result:\n", nemenyi_result)
print("Nemenyi df:\n", nemenyi_df)
print("Nemenyi table:\n", nemenyi_table)
print("Nemenyi df 2:\n", nemenyi_df_2)
print("Nemenyi table 2:\n", nemenyi_table_2)






import glob
import pickle
import os

def find_max_value(dictionary, key_to_find, exclude_key=None):
    max_value = float('-inf')
    for key, value in dictionary.items():
        if key == exclude_key:
            continue
        if key == key_to_find:
            max_value = max(max_value, value)
        if isinstance(value, dict):
            max_value = max(max_value, find_max_value(value, key_to_find, exclude_key))
        elif isinstance(value, list) and isinstance(value[0], dict):
            for item in value:
                max_value = max(max_value, find_max_value(item, key_to_find, exclude_key))
    return max_value

def find_max_values_in_files(directory):
    max_values = {}
    for filepath in glob.glob(directory + '**/*.pkl', recursive=True):
        with open(filepath, 'rb') as file:
            data = pickle.load(file)
            relative_path = os.path.relpath(filepath, directory)
            directory_name = os.path.dirname(relative_path)
            key = directory_name.replace(os.path.sep, '_') + '_maximum'
            max_values[key] = find_max_value(data, 'n_inliers_maximum', 'standard_RANSAC_100000')
    return max_values

directory = "/home/scpmaotj/Stanford3dDataset_v1.2/"  # specify your directory
max_values = find_max_values_in_files(directory)

# Save the results to disk
with open('max_values.pkl', 'wb') as file:
    pickle.dump(max_values, file)





import mrdja.ransac.coreransac as coreransac
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geometry
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

def get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed):
    dict_full_results = {}
    dict_full_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    '''
    audit_before_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_before_sanitizing"] = audit_before_sanitizing
    pcd = pointcloud.pointcloud_sanitize(pcd)
    audit_after_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_after_sanitizing"] = audit_after_sanitizing
    '''
    number_pcd_points = len(pcd.points)
    dict_full_results["number_pcd_points"] = number_pcd_points
    np_points = np.asarray(pcd.points)

    dict_full_results["ransac_iterations_results"] = []

    np.random.seed(seed)
    max_number_inliers = 0
    best_iteration_results = None
    for _ in range(ransac_iterations):
        dict_iteration_results = coreransac.get_ransac_iteration_results(np_points, threshold, number_pcd_points)
        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results

ransac_iterations = 1000
threshold = 0.002
seed = 42

filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
RANSAC_data_from_file = get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed)

print(RANSAC_data_from_file)

# create a np array with all the planes
list_iterations_results = RANSAC_data_from_file["ransac_iterations_results"]
all_planes = np.empty((0, 4))
for iteration_results in list_iterations_results:
    current_plane = iteration_results["current_plane"]
    # convert current_plane shape from (4,) to (1, 4)
    current_plane = np.expand_dims(current_plane, axis=0)
    print(current_plane)
    print(current_plane.shape)
    print(all_planes.shape)
    all_planes = np.append(all_planes, current_plane, axis=0)
print(all_planes.shape)

# create a list with all the numbers of inliers
list_number_inliers = []
for iteration_results in list_iterations_results:
    list_number_inliers.append(iteration_results["number_inliers"])
print(list_number_inliers)

# plot a histogram of the number of inliers
import matplotlib.pyplot as plt
plt.hist(list_number_inliers, bins=100)
plt.show()

import mrdja.ransac.coreransacutils as crsu
inliers_ratio = list_number_inliers[0] / RANSAC_data_from_file["number_pcd_points"]
number_iterations = crsu.compute_number_iterations(inliers_ratio, alpha=0.95)
print(number_iterations)

# get the maximum number of inliers and the position of the iteration that produced it
max_number_inliers = 0
max_number_inliers_index = 0
for index, number_inliers in enumerate(list_number_inliers):
    if number_inliers > max_number_inliers:
        max_number_inliers = number_inliers
        max_number_inliers_index = index

# compute the number of iterations needed to get a probability of 0.95 of finding the best model
inliers_ratio = max_number_inliers / RANSAC_data_from_file["number_pcd_points"]
number_iterations = crsu.compute_number_iterations(inliers_ratio, alpha=0.99)
print(number_iterations)

# extract normalized normals from the planes, and also the corresponding D parameter after normalization
all_normals = all_planes[:, 0:3]
print(all_normals.shape)
all_D = all_planes[:, 3]
all_D = all_D / np.linalg.norm(all_normals, axis=1)
all_normals = all_normals / np.linalg.norm(all_normals, axis=1)[:, None]
print(all_normals.shape)

# plot the normals as points in 3D space using matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')

x = all_normals[:, 0]
y = all_normals[:, 1]
z = all_normals[:, 2]

ax.scatter(x, y, z, c='r', marker='o')

ax.set_xlabel('X Label')
ax.set_ylabel('Y Label')
ax.set_zlabel('Z Label')

plt.show()

# plot a histogram of all_D values
plt.hist(all_D, bins=100)
plt.show()










import mrdja.ransac.coreransac as coreransac
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geometry
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

def get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed):
    dict_full_results = {}
    dict_full_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    '''
    audit_before_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_before_sanitizing"] = audit_before_sanitizing
    pcd = pointcloud.pointcloud_sanitize(pcd)
    audit_after_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_after_sanitizing"] = audit_after_sanitizing
    '''
    number_pcd_points = len(pcd.points)
    dict_full_results["number_pcd_points"] = number_pcd_points
    np_points = np.asarray(pcd.points)

    dict_full_results["ransac_iterations_results"] = []

    np.random.seed(seed)
    max_number_inliers = 0
    best_iteration_results = None
    for _ in range(ransac_iterations):
        dict_iteration_results = coreransac.get_ransac_iteration_results(np_points, threshold, number_pcd_points)
        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results

ransac_iterations = 800
threshold = 0.02
seed = 42

filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
RANSAC_data_from_file = get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed)

print(RANSAC_data_from_file)

# create a function that takes the RANSAC_data_from_file and the iteration number and returns the inliers
# create a function that takes the RANSAC_data_from_file and the iteration number and returns the number of inliers

pcd = o3d.io.read_point_cloud(RANSAC_data_from_file["filename"])
np_points = np.asarray(pcd.points)

def get_inliers_from_iteration(RANSAC_data_from_file, np_points, iteration_number):
    dict_iteration_results = RANSAC_data_from_file["ransac_iterations_results"][iteration_number]
    inliers = np_points[dict_iteration_results["indices_inliers"]]
    pcd_inliers = o3d.geometry.PointCloud()
    pcd_inliers.points = o3d.utility.Vector3dVector(inliers)
    pcd_inliers.paint_uniform_color([1, 0, 0])
    return pcd_inliers

def get_number_inliers_from_iteration(RANSAC_data_from_file, iteration_number):
    dict_iteration_results = RANSAC_data_from_file["ransac_iterations_results"][iteration_number]
    return dict_iteration_results["number_inliers"]

# using open3d create a geometry with the inliers painted red, and show the inliers, the original point cloud and the number of iteration and the number
# of inliers in the same window; create a for loop that shows the inliers for each iteration

def draw_inliers_from_iteration(RANSAC_data_from_file, pcd, iteration_number):
    pcd_inliers = get_inliers_from_iteration(RANSAC_data_from_file, np_points, iteration_number)
    number_inliers = get_number_inliers_from_iteration(RANSAC_data_from_file, iteration_number)
    o3d.visualization.draw_geometries([pcd, pcd_inliers], window_name="Inliers from iteration " + str(iteration_number) + " with " + str(number_inliers) + " inliers")

max_number_inliers = 0
for iteration_number in range(ransac_iterations):
    # only show if the number of inliers is greater than threshold
    threshold_inliers = 40000
    current_number_inliers = get_number_inliers_from_iteration(RANSAC_data_from_file, iteration_number)
    if current_number_inliers > max_number_inliers:
        max_number_inliers = current_number_inliers
    if current_number_inliers > threshold_inliers:
        draw_inliers_from_iteration(RANSAC_data_from_file, pcd, iteration_number)

print("max_number_inliers: " + str(max_number_inliers))







import mrdja.ransaclp as ransaclp
import mrdja.pointcloud as pointcloud
from typing import Dict, List
import numpy as np
import open3d as o3d
import glob
import os

def get_baseline(filename: str, threshold: float, n_iterations = 100000) -> Dict:
    '''
    Get the open3d results for plane segmentation for a high number of iterations.
    
    :param filename: The path to the file to be processed.
    :type filename: str
    :param threshold: The threshold to be used in the RANSAC algorithm.
    :type threshold: float
    :param n_iterations: The number of iterations to be used in the RANSAC algorithm.
    :type n_iterations: int
    :return: A dictionary with the results.
    :rtype: Dict

    :Example:

    ::

        >>> import mrdja.ransaclpexperiments as experiments
        >>> filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"
        >>> threshold = 0.02
        >>> n_iterations = 100000
        >>> dict_results = experiments.get_baseline(filename, threshold, n_iterations)
        >>> print(dict_results)
        {'plane_model': array([-0.00485739, -0.00538778,  0.99997369,  0.10760573]), 'number_inliers': 154556}
    
    '''
    pcd = o3d.io.read_point_cloud(filename)
    plane_model, inliers = pcd.segment_plane(distance_threshold=threshold,
                                            ransac_n=3,
                                            num_iterations=n_iterations)
    return {"plane_model": plane_model, "number_inliers": len(inliers)}

def get_baseline_S3DIS(database_path: str, threshold: float, n_iterations = 100000) -> Dict:
    '''
    Get the open3d results for plane segmentation in the S3DIS dataset for a high number of iterations.

    :param database_path: The path to the S3DIS dataset.
    :type database_path: str
    :param threshold: The threshold to be used in the RANSAC algorithm.
    :type threshold: float
    :param n_iterations: The number of iterations to be used in the RANSAC algorithm.
    :type n_iterations: int
    :return: A dictionary with the results.
    :rtype: Dict

    :Example:

    ::

        >>> import mrdja.ransaclpexperiments as experiments
        >>> import pickle as pkl 
        >>> database_path = "/home/scpmaotj/Stanford3dDataset_v1.2/"
        >>> threshold = 0.02
        >>> n_iterations = 100000
        >>> dict_results = experiments.get_baseline_S3DIS(database_path, threshold, n_iterations)
        >>> dict_results['/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/office_29/office_29.ply']
        {'plane_model': array([-0.00794638,  0.00484077,  0.99995671, -2.77513434]),
        'number_inliers': 71578}
        >>> # save results to a pickle file
        >>> with open('results_baseline_S3DIS.pkl', 'wb') as f:
        ...     pkl.dump(dict_results, f)


    '''
    ply_files = glob.glob(database_path + "/**/*.ply", recursive=True)
    dict_all_results = {}

    for filename in ply_files:
        current_file_baseline = get_baseline(filename, threshold = threshold, n_iterations = n_iterations)
        dict_all_results[filename] = current_file_baseline

    return dict_all_results

def compute_parameters_ransac_line (line_iterations: int, percentage_chosen_lines: float = 0.2, percentage_chosen_planes: float = 0.05) -> Dict:
    '''
    Compute the parameters for the RANSAC line algorithm from the number of iterations.

    :param line_iterations: The number of iterations to be used in the RANSAC line algorithm.
    :type line_iterations: int
    :return: A tuple with the number of chosen lines, the number of line pairs, the number of chosen planes and the total number of iterations.
    :rtype: Dict

    :Example:

    ::

        >>> import mrdja.ransaclpexperiments as experiments
        >>> line_iterations = 200
        >>> experiments.compute_parameters_ransac_line(line_iterations)
        {'number_chosen_lines': 40,
        'number_lines_pairs': 780,
        'number_chosen_planes': 39,
        'total_iterations': 239}
        >>> experiments.compute_parameters_ransac_line(line_iterations, percentage_chosen_lines = 0.1, percentage_chosen_planes = 0.1)
        {'number_chosen_lines': 20,
        'number_lines_pairs': 190,
        'number_chosen_planes': 19,
        'total_iterations': 219}
    '''
    number_chosen_lines = int(line_iterations * percentage_chosen_lines)
    number_lines_pairs = int (number_chosen_lines * (number_chosen_lines - 1) / 2)
    number_chosen_planes = int(number_lines_pairs * percentage_chosen_planes)
    total_iterations = line_iterations + number_chosen_planes
    return {"number_chosen_lines": number_chosen_lines, "number_lines_pairs": number_lines_pairs, "number_chosen_planes": number_chosen_planes, "total_iterations": total_iterations}

def get_data_comparison_ransac_and_ransaclp(filename: str, repetitions: int, iterations_list: List[int], threshold: float, 
                                            percentage_chosen_lines: float, percentage_chosen_planes: float, 
                                            cuda: bool = False, verbosity_level: int = 0, 
                                            inherited_verbose_string: str = "",
                                            seed: int = None) -> Dict :
    '''
    Get the data for the comparison between RANSAC and RANSACLP.

    :param filename: The path to the file to be processed.
    :type filename: str
    :param repetitions: The number of repetitions to be used.
    :type repetitions: int
    :param iterations_list: The list of iterations to be used.
    :type iterations_list: List[int]
    :param threshold: The threshold to be used in the RANSAC algorithm.
    :type threshold: float
    :param percentage_chosen_lines: The percentage of chosen lines to be used in the RANSAC line algorithm.
    :type percentage_chosen_lines: float
    :param percentage_chosen_planes: The percentage of chosen planes to be used in the RANSAC line algorithm.
    :type percentage_chosen_planes: float
    :param verbosity_level: The verbosity level to be used.
    :type verbosity_level: int
    :param inherited_verbose_string: The inherited verbose string to be used.
    :type inherited_verbose_string: str
    :param seed: The seed to be used.
    :type seed: int
    :return: A dictionary with the results.
    :rtype: Dict
    '''
    if iterations_list is None:
        raise ValueError("iterations_list cannot be None")
    if len(iterations_list) == 0:
        raise ValueError("iterations_list cannot be empty")
    if repetitions <= 0:
        raise ValueError("repetitions must be greater than 0")
    if percentage_chosen_lines <= 0 or percentage_chosen_lines > 1:
        raise ValueError("percentage_chosen_lines must be greater than 0 and less or equal than 1")
    if percentage_chosen_planes <= 0 or percentage_chosen_planes > 1:
        raise ValueError("percentage_chosen_planes must be greater than 0 and less or equal than 1")
    
    if seed is not None:
        np.random.seed(seed)
    
    dict_all_results = {}
    dict_all_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    pcd = pointcloud.pointcloud_sanitize(pcd)
    np_points = np.asarray(pcd.points)

    dict_all_results["number_pcd_points"] = len(np_points)
    dict_all_results["threshold"] = threshold

    # get the maximum number of iterations
    max_iterations = max(iterations_list)
    filename_only_file = os.path.basename(filename)
    inherited_verbose_string = inherited_verbose_string + filename_only_file + " "

    for index, num_iterations in enumerate(iterations_list):
        inherited_verbose_string_in_first_loop = f"{inherited_verbose_string} Current max RANSAC iterations {num_iterations} {index+1}/{len(iterations_list)}"
        if verbosity_level > 0:
            print(f"{inherited_verbose_string_in_first_loop} Current number of iterations analyzed: {num_iterations} / {max_iterations}")
        parameters_experiment = compute_parameters_ransac_line(num_iterations, percentage_chosen_lines = percentage_chosen_lines, 
                                                               percentage_chosen_planes = percentage_chosen_planes)
        total_iterations = parameters_experiment["total_iterations"]

        dict_standard_RANSAC_results_list = list()
        dict_line_RANSAC_results_list = list()

        current_repetition = 0
        for j in range(repetitions):
            current_repetition = current_repetition + 1
            inherited_verbose_string_in_second_loop = f"{inherited_verbose_string_in_first_loop} Repetition {current_repetition}/{repetitions} "
            ransaclp_data_from_file = ransaclp.get_ransaclp_data_from_np_points(np_points, ransac_iterations = num_iterations, 
                                                           threshold = threshold,
                                                           cuda = cuda,
                                                           verbosity_level = verbosity_level, 
                                                           inherited_verbose_string = inherited_verbose_string_in_second_loop,
                                                           seed = None)
            ransaclp_number_inliers = ransaclp_data_from_file["number_inliers"]
            ransaclp_plane = ransaclp_data_from_file["plane"]
            
            ransac_plane, inliers = pcd.segment_plane(distance_threshold=threshold,
                                                    ransac_n=3,
                                                    num_iterations=total_iterations)
            ransac_number_inliers = len(inliers)

            dict_standard_RANSAC_results = {"number_inliers": ransac_number_inliers, "plane": ransac_plane, "plane_iterations": total_iterations}
            dict_line_RANSAC_results = {"number_inliers": ransaclp_number_inliers, "plane": ransaclp_plane, "line_iterations": num_iterations}
            dict_line_RANSAC_results_list.append(dict_line_RANSAC_results)
            dict_standard_RANSAC_results_list.append(dict_standard_RANSAC_results)

        dict_all_results["standard_RANSAC_" + str(total_iterations)] = dict_standard_RANSAC_results_list
        dict_all_results["line_RANSAC_" + str(total_iterations)] = dict_line_RANSAC_results_list
        # get the mean of n_inliers_maximum of the elements of the list dict_line_RANSAC_results_list
        list_n_inliers_maximum = [int(dict_line_RANSAC_results["number_inliers"]) for dict_line_RANSAC_results in dict_line_RANSAC_results_list]
        mean_n_inliers_maximum = np.mean(list_n_inliers_maximum)
        dict_all_results["mean_number_inliers_line_RANSAC_" + str(total_iterations)] = mean_n_inliers_maximum
        # get the mean of n_inliers_maximum of the elements of the list dict_standard_RANSAC_results_list
        list_n_inliers_maximum = [int(dict_standard_RANSAC_results["number_inliers"]) for dict_standard_RANSAC_results in dict_standard_RANSAC_results_list]
        mean_n_inliers_maximum = np.mean(list_n_inliers_maximum)
        dict_all_results["mean_number_inliers_standard_RANSAC_" + str(total_iterations)] = mean_n_inliers_maximum

    return dict_all_results



import mrdja.ransaclpexperiments as experiments
import open3d as o3d

dataset = o3d.data.OfficePointClouds()
office_paths = dataset.paths
filename = office_paths[0]

threshold = 0.02
# repetitions = 10
repetitions = 3
# iterations_list = [100, 200, 300, 400, 500, 600]
iterations_list = [100, 200]
percentage_chosen_lines = 0.2
percentage_chosen_planes = 0.05
seed = 42

dict_results = experiments.get_data_comparison_ransac_and_ransaclp(filename = filename, repetitions = repetitions, 
                                                                   iterations_list = iterations_list, threshold = threshold, 
                                                                    percentage_chosen_lines = percentage_chosen_lines, 
                                                                    percentage_chosen_planes = percentage_chosen_planes,
                                                                    cuda = True, 
                                                                    verbosity_level = 1, 
                                                                    seed = seed)

print(dict_results)


import os
import pickle

directory = "/home/scpmaotj/pkl_files/"  # Replace with the actual directory path

for filename in os.listdir(directory):
    if filename.endswith(".pkl"):
        old_filepath = os.path.join(directory, filename)
        new_filename = f"{os.path.splitext(filename)[0]}_means.pkl"
        new_filepath = os.path.join(directory, new_filename)

        with open(old_filepath, 'rb') as file:
            old_dict = pickle.load(file)

        new_dict = {
            key: value for key, value in old_dict.items()
            if 'mean' in key
        }

        with open(new_filepath, 'wb') as file:
            pickle.dump(new_dict, file)

        print(f"Processed file: {filename} --> Created {new_filename}")

print("All files processed and saved.")



import open3d as o3d
import numpy as np
import mrdja.geometry as geometry
import mrdja.sampling as sampling
import mrdja.drawing as drawing
import mrdja.ransac.coreransac as coreransac
import mrdja.ransac as ransac
from sklearn.cluster import DBSCAN, KMeans

# Create a 3D parallelogram
n_samples = 10000
normal1 = (1, 0, 0)
normal2 = (0, 1, 0)
normal3 = (0, 0, 1)
center = (0, 0, 0)
length1 = 10
length2 = 10
length3 = 0.5
normal1 = np.array(normal1)
normal2 = np.array(normal2)
normal3 = np.array(normal3)
center = np.array(center)
samples = sampling.sampling_parallelogram_3d(n_samples, normal1, normal2, normal3, center, length1, length2, length3)
samples = np.array(samples)

# Create a 3D point cloud
pcd_plane = o3d.geometry.PointCloud()
pcd_plane.points = o3d.utility.Vector3dVector(samples)
pcd_plane.paint_uniform_color([0, 1, 0])

# Get the plane equation
plane_equation = geometry.get_plane_equation(normal1, normal2, center)

sphere_radius = 10
sphere_center = (0, 0, 0)
sphere_points = sampling.sampling_sphere(n_samples=100, center=sphere_center, radius=sphere_radius, seed=None)
# Create a 3D point cloud
pcd_sphere_points = o3d.geometry.PointCloud()
pcd_sphere_points.points = o3d.utility.Vector3dVector(sphere_points)
pcd_sphere_points.paint_uniform_color([1, 0, 0])

mesh_sphere = o3d.geometry.TriangleMesh.create_sphere(radius=sphere_radius, resolution=20) # created at origin
mesh_sphere = mesh_sphere.translate(sphere_center)
pcd_sphere_boundary = mesh_sphere.sample_points_poisson_disk(number_of_points=100, init_factor=5)
pcd_sphere_boundary.paint_uniform_color([1, 0, 0])

# Draw the point cloud, the plane and the sphere
true_plane_geometry = drawing.draw_plane_as_lines_open3d(*plane_equation, line_color=[0, 0, 1])
o3d.visualization.draw_geometries([pcd_plane, pcd_sphere_points, pcd_sphere_boundary, true_plane_geometry])
list_geometries = [pcd_plane, pcd_sphere_points, pcd_sphere_boundary, true_plane_geometry]

pcd = pcd_plane + pcd_sphere_points + pcd_sphere_boundary

len_points = len(pcd.points)
for i in range(20):
    ransac_results = coreransac.get_ransac_iteration_results(np.asanyarray(pcd.points), len_points, threshold=0.5)
    print(ransac_results)
    current_plane = ransac_results["current_plane"]
    list_geometries.append(drawing.draw_plane_as_lines_open3d(*current_plane, line_color=[0, 1, 1]))
o3d.visualization.draw_geometries(list_geometries)

# get 20 random points from the pcd_plane point cloud
random_points = sampling.sampling_pcd_points(pcd_plane, 200)
closest_plane = geometry.find_closest_plane(random_points)
# draw the closest plane along with the random points
closest_plane_geometry = drawing.draw_plane_as_lines_open3d(*closest_plane, line_color=[0, 1, 1])
o3d.visualization.draw_geometries([pcd_plane, closest_plane_geometry, true_plane_geometry])

k = 500
dict_distances = ransac.generate_k_planes_from_k_sets_of_3_random_points_of_pointcloud_points_and_compute_the_distance_from_the_planes_to_all_the_points(pcd, k)
# distances is a dictionary of the form {plane: distances}
# where plane is a tuple of the form (A, B, C, D) and distances is a np.array of shape (len(pcd.points),)
# apply KMEANS clustering to the distances
distances = np.array(list(dict_distances.values()))
planes = np.array(list(dict_distances.keys()))
kmeans = KMeans(n_clusters=5, random_state=0).fit(distances)
# get the labels
labels = kmeans.labels_
# get the number of clusters
n_clusters = len(set(labels)) - (1 if -1 in labels else 0)
# get the indices of the points that belong to each cluster
clusters_indices = [np.where(labels == i)[0] for i in range(n_clusters)]
# get the planes that belong to each cluster
clusters_planes = []
for indices_list in clusters_indices:
    print(indices_list)
    clusters_planes.append(planes[indices_list])
# plot the planes that belong to each cluster in a different color
list_geometries = [pcd_plane, true_plane_geometry]
for cluster_plane in clusters_planes:
    line_color = np.random.rand(3)
    for plane in cluster_plane:
        list_geometries.append(drawing.draw_plane_as_lines_open3d(*plane, line_color=line_color))
o3d.visualization.draw_geometries(list_geometries)
# order the clusters_planes by the number of planes they contain
clusters_planes = sorted(clusters_planes, key=lambda x: len(x), reverse=True)
list_geometries = [pcd_plane, true_plane_geometry]
for plane in clusters_planes[0]:
    list_geometries.append(drawing.draw_plane_as_lines_open3d(*plane, line_color=[0, 1, 1]))
o3d.visualization.draw_geometries(list_geometries)



import numpy as np
import math

def first_version(point, B, C):
    cross_product = np.cross(point - B, C - B)
    magnitude_cross_product = np.linalg.norm(cross_product)
    magnitude_C_minus_B = np.linalg.norm(C - B)
    distance = magnitude_cross_product / magnitude_C_minus_B
    return distance

def second_version(point, B, C):
    V = point - B
    W = C - B
    cross_product_x = V[1] * W[2] - V[2] * W[1]
    cross_product_y = V[2] * W[0] - V[0] * W[2]
    cross_product_z = V[0] * W[1] - V[1] * W[0]
    
    magnitude_cross_product = math.sqrt(cross_product_x * cross_product_x + cross_product_y * cross_product_y + cross_product_z * cross_product_z)
    magnitude_C_minus_B = math.sqrt((C[0] - B[0]) * (C[0] - B[0]) + (C[1] - B[1]) * (C[1] - B[1]) + (C[2] - B[2]) * (C[2] - B[2]))
    
    dist = magnitude_cross_product / magnitude_C_minus_B
    return dist

# test if the two versions are the same
np.random.seed(42)
# create a list with dist1 and dist2 for every point
dist1_list = []
dist2_list = []
for _ in range(100):
    point = np.random.rand(3)
    B = np.random.rand(3)
    C = np.random.rand(3)
    dist1 = first_version(point, B, C)
    dist2 = second_version(point, B, C)
    dist1_list.append(dist1)
    dist2_list.append(dist2)
dist1_list = np.array(dist1_list)
dist2_list = np.array(dist2_list)
print(np.allclose(dist1_list, dist2_list))


import gradio as gr
import mrdja.sampling as sampling
import mrdja.geometry as geometry
import matplotlib.pyplot as plt
from gradio.components import Number, Image
import matplotlib
import numpy as np
matplotlib.use('Agg')

def plot_circle(n_samples, center_x, center_y, radius):
    center = (center_x, center_y)
    samples = sampling.sampling_circle_2d(n_samples=n_samples, center=center, radius=radius)
    fig, ax = plt.subplots()
    ax.scatter(*zip(*samples))
    ax.set_aspect('equal')
    xlim_min = center_x - radius
    xlim_max = center_x + radius
    ylim_min = center_y - radius
    ylim_max = center_y + radius
    graph_limits = geometry.get_limits_of_graph_from_limits_of_object(xlim_min, xlim_max, ylim_min, ylim_max)
    ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    ax.set_ylim(graph_limits[2], graph_limits[3])  # Set y-axis limits
    # create title from n_samples, center, and radius, using f-string
    ax.set_title(f'{n_samples} Samples on a Circle with Center {center} and Radius {radius}')
    # draw also the circle in red
    circle = plt.Circle(center, radius, color='r', fill=False)
    ax.add_artist(circle)

    # Draw the X and Y axes in dotted lines
    ax.axhline(0, linestyle='dotted', color='black')
    ax.axvline(0, linestyle='dotted', color='black')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    fig.canvas.draw()
    img = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close(fig)
    return img

# Define the input and output interfaces for Gradio
inputs = [
    Number(value=100, precision=0, label="Number of Samples"),
    Number(value=2, label="Center X-coordinate"),
    Number(value=3, label="Center Y-coordinate"),
    Number(value=5, label="Radius")
]
outputs = [
    Image(label="Circle Plot", type="pil")
]

# Create the Gradio app
app = gr.Interface(
    fn=plot_circle, 
    inputs=inputs, 
    outputs=outputs,
    title="Sampling 2D Circles",
    description="Generate random samples on a 2D circle with a specified radius and center.",
    allow_flagging="never",
)

# Run the app
app.launch()



'''
Module for functions that take an ax object, draw something, and return the ax object. This is useful for chaining calls to these functions.
'''

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
import mrdja.geometry as geom
import numpy as np

def draw_segment(line: np.ndarray, color: str = 'red', style: str = '-', ax = None):
    '''
    Draws a line in 3D space.

    :param segment: A 2x3 numpy array containing the endpoints of the segment.
    :type segment: numpy.ndarray
    :param color: The color of the segment to draw.
    :type color: str
    :param style: The style of the segment to draw.
    :type style: str
    :param ax: The matplotlib ax object to draw on.
    :type ax: matplotlib.axes._subplots.Axes3DSubplot
    :return: The ax object.
    :rtype: matplotlib.axes._subplots.Axes3DSubplot

    :Example:

    ::

    >>> import mrdja.matplot3d as plt3d
    >>> import numpy as np
    >>> segment = np.array([[0, 0, 0], [1, 1, 1]])
    >>> plt3d.draw_segment(segment, color="blue", style="--")
    
    |matplot3d_draw_segment_example|

    .. |matplot3d_draw_segment_example| image:: ../../_static/images/matplot3d_draw_segment_example.png
    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    ax.plot(line[:, 0], line[:, 1], line[:, 2], color=color, linestyle=style)
    if plot_inside_function:
        plt.show()
    else:
        return ax
    

def draw_polygon(vertices: np.ndarray, color: str = 'red', alpha: float = 0.5, ax = None):
    '''
    Draws a polygon in 3D space.

    :param vertices: A Nx3 numpy array containing the vertices of the polygon.
    :type vertices: numpy.ndarray
    :param color: The color of the polygon to draw.
    :type color: str
    :param style: The style of the polygon to draw.
    :type style: str
    :param ax: The matplotlib ax object to draw on.
    :type ax: matplotlib.axes._subplots.Axes3DSubplot
    :return: The ax object.
    :rtype: matplotlib.axes._subplots.Axes3DSubplot

    :Example:

    ::

    >>> import mrdja.matplot3d as plt3d
    >>> import numpy as np
    >>> vertices = np.array([[0, 0, 0], [1, 1, 1], [1, 0, 0]])
    >>> plt3d.draw_polygon(vertices, color="blue", alpha=0.2)
    
    |matplot3d_draw_polygon_example|

    .. |matplot3d_draw_polygon_example| image:: ../../_static/images/matplot3d_draw_polygon_example.png
    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    # Create a collection of polygons
    polygon = [vertices]
    # Plot the polygon
    ax.add_collection3d(Poly3DCollection(polygon, facecolors=color, alpha=alpha))

    if plot_inside_function:
        plt.show()
    else:
        return ax
    
def draw_circle(center: np.ndarray = np.array([0, 0, 0]), radius: float = 1, normal: np.ndarray = np.array([0, 0 ,1]), color: str = 'red', alpha: float = 0.5, ax = None):
    '''
    Draws a circle in 3D space.

    :param center: A 3x1 numpy array containing the center of the circle.
    :type center: numpy.ndarray
    :param radius: The radius of the circle to draw.
    :type radius: float
    :param normal: The normal vector of the circle.
    :type normal: numpy.ndarray
    :param color: The color of the circle to draw.
    :type color: str
    :param style: The style of the circle to draw.
    :type style: str
    :param ax: The matplotlib ax object to draw on.
    :type ax: matplotlib.axes._subplots.Axes3DSubplot
    :return: The ax object.
    :rtype: matplotlib.axes._subplots.Axes3DSubplot

    :Example:

    ::

    >>> import mrdja.matplot3d as plt3d
    >>> import numpy as np
    >>> center = np.array([0, 0, 0])
    >>> radius = 1
    >>> normal = np.array([0.5, 0.3, 0.8]) / np.linalg.norm(np.array([0.5, 0.3, 0.8]))
    >>> plt3d.draw_circle(center, radius, normal, color="blue", alpha=0.2)

    |matplot3d_draw_circle_example|

    .. |matplot3d_draw_circle_example| image:: ../../_static/images/matplot3d_draw_circle_example.png

    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    # Generate points on the circle's surface
    num_points = 100
    theta = np.linspace(0, 2 * np.pi, num_points)
    phi = np.linspace(0, np.pi, num_points)
    theta, phi = np.meshgrid(theta, phi)

    # Calculate an orthogonal vector to the given normal vector
    orthogonal_vector = np.array([1, 0, 0]) if normal[1] == 0 else np.array([0, 1, 0])

    # Calculate the third orthogonal vector
    third_vector = np.cross(normal, orthogonal_vector)

    # Calculate the points on the filled circle
    x = center[0] + radius * (orthogonal_vector[0] * np.sin(phi) * np.cos(theta) + third_vector[0] * np.sin(phi) * np.sin(theta))
    y = center[1] + radius * (orthogonal_vector[1] * np.sin(phi) * np.cos(theta) + third_vector[1] * np.sin(phi) * np.sin(theta))
    z = center[2] + radius * (orthogonal_vector[2] * np.sin(phi) * np.cos(theta) + third_vector[2] * np.sin(phi) * np.sin(theta))

    # Plot the filled circle
    ax.add_collection3d(Poly3DCollection([list(zip(x.flatten(), y.flatten(), z.flatten()))], facecolors=color, alpha=alpha, linewidths=0))

    if plot_inside_function:
        # Compute the minimum and maximum values of x
        min_x = np.min(x)
        max_x = np.max(x)
        # Compute the minimum and maximum values of y
        min_y = np.min(y)
        max_y = np.max(y)
        # Compute the minimum and maximum values of z
        min_z = np.min(z)
        max_z = np.max(z)
        # Compute the limits of the 3D graph
        graph_limits = geom.get_limits_of_3d_graph_from_limits_of_object(min_x, max_x, min_y, max_y, min_z, max_z)
        ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
        ax.set_ylim(graph_limits[2], graph_limits[3])  # Set y-axis limits
        ax.set_zlim(graph_limits[4], graph_limits[5])  # Set z-axis limits
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        plt.show()
    else:
        return ax

def draw_circumference(center: np.ndarray = np.array([0, 0, 0]), radius: float = 1, normal: np.ndarray = np.array([0, 0 ,1]), color: str = 'red', alpha: float = 0.5, ax = None):
    '''
    Draws a circumference in 3D space.

    :param center: A 3x1 numpy array containing the center of the circumference.
    :type center: numpy.ndarray
    :param radius: The radius of the circumference to draw.
    :type radius: float
    :param normal: The normal vector of the circumference.
    :type normal: numpy.ndarray
    :param color: The color of the circumference to draw.
    :type color: str
    :param style: The style of the circumference to draw.
    :type style: str
    :param ax: The matplotlib ax object to draw on.
    :type ax: matplotlib.axes._subplots.Axes3DSubplot
    :return: The ax object.
    :rtype: matplotlib.axes._subplots.Axes3DSubplot

    :Example:

    ::

    >>> import mrdja.matplot3d as plt3d
    >>> import numpy as np
    >>> center = np.array([0, 0, 0])
    >>> radius = 1
    >>> normal = np.array([0.5, 0.3, 0.8]) / np.linalg.norm(np.array([0.5, 0.3, 0.8]))
    >>> plt3d.draw_circumference(center, radius, normal, color="blue", alpha=0.2)

    |matplot3d_draw_circumference_example|

    .. |matplot3d_draw_circumference_example| image:: ../../_static/images/matplot3d_draw_circumference_example.png

    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    # Generate points on a 2D circle in the XY plane
    theta = np.linspace(0, 2 * np.pi, 100)
    x = radius * np.cos(theta)
    y = radius * np.sin(theta)

    # Calculate an orthogonal vector to the normal vector
    orthogonal_vector = np.array([1, 0, 0]) if normal[1] == 0 else np.array([0, 1, 0])

    # Calculate the third orthogonal vector
    third_vector = np.cross(normal, orthogonal_vector)

    # Apply a transformation to rotate and translate the circle
    transformation_matrix = np.column_stack((orthogonal_vector, third_vector, normal))
    transformed_points = np.dot(transformation_matrix, np.vstack((x, y, np.zeros_like(x))))
    translated_points = center[:, np.newaxis] + transformed_points

    # Plot the 3D circle
    ax.plot(translated_points[0], translated_points[1], translated_points[2], color=color, alpha=alpha)

    if plot_inside_function:
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        plt.show()
    else:
        return ax
    
def draw_points(points: np.ndarray, color: str = 'red', style: str = "o", alpha: float = 0.5, ax = None):
    '''
    Draws points in 3D space.

    :param points: A Nx3 numpy array containing the points to draw.
    :type points: numpy.ndarray
    :param color: The color of the points to draw.
    :type color: str
    :param style: The style of the points to draw.
    :type style: str
    :param alpha: The transparency of the points to draw.
    :type alpha: float
    :param ax: The matplotlib ax object to draw on.
    :type ax: matplotlib.axes._subplots.Axes3DSubplot
    :return: The ax object.
    :rtype: matplotlib.axes._subplots.Axes3DSubplot

    :Example:

    ::

    >>> import mrdja.matplot3d as plt3d
    >>> import numpy as np
    >>> points = np.array([[0, 0, 0], [1, 1, 1], [1, 0, 0]])
    >>> plt3d.draw_points(points, color="blue", style="o")


    |matplot3d_draw_points_example|

    .. |matplot3d_draw_points_example| image:: ../../_static/images/matplot3d_draw_points_example.png
    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    ax.scatter(points[:, 0], points[:, 1], points[:, 2], color=color, marker=style, alpha=alpha)
    if plot_inside_function:
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        plt.show()
    else:
        return ax
